/*
 * Copyright (c) 2015 Kaprica Security, Inc.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 */
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "kx.h"

#include "kxdata.h"

void kx_init_common(kx_t *kx)
{
    bn_init(&kx->P);
    bn_init(&kx->Q);
    bn_init(&kx->G);
    bn_init(&kx->a);
    bn_init(&kx->b);
    bn_init(&kx->S);
}

int kx_init_std(kx_t *kx, unsigned int group)
{
    if (group >= KX_GROUP_END)
        return FAILURE;

    kx_init_common(kx);

    if (bn_from_bytes(&kx->P, groups[group].P, sizeof(groups[group].P)) != SUCCESS)
        goto fail;
    if (bn_from_bytes(&kx->Q, groups[group].Q, sizeof(groups[group].Q)) != SUCCESS)
        goto fail;
    if (bn_from_bytes(&kx->G, groups[group].G, sizeof(groups[group].G)) != SUCCESS)
        goto fail;
    return SUCCESS;

fail:
    bn_destroy(&kx->P);
    bn_destroy(&kx->Q);
    bn_destroy(&kx->G);
    return FAILURE;
}

int kx_init(kx_t *kx, bn_t *P, bn_t *Q, bn_t *G)
{
    kx_init_common(kx);

    if (bn_copy(&kx->P, P) != SUCCESS)
        goto fail;
    if (bn_copy(&kx->Q, Q) != SUCCESS)
        goto fail;
    if (bn_copy(&kx->G, G) != SUCCESS)
        goto fail;
    return SUCCESS;

fail:
    bn_destroy(&kx->P);
    bn_destroy(&kx->Q);
    bn_destroy(&kx->G);
    return FAILURE;
}

void kx_destroy(kx_t *kx)
{
    bn_destroy(&kx->P);
    bn_destroy(&kx->Q);
    bn_destroy(&kx->G);
    bn_destroy(&kx->a);
    bn_destroy(&kx->b);
    bn_destroy(&kx->S);
}

int kx_gen_a(kx_t *kx, rng_t *rng)
{
    return bn_random(&kx->a, &kx->Q, rng);
}

int kx_get_A(kx_t *kx, bn_t *A)
{
    return bn_modexp(A, &kx->G, &kx->a, &kx->P);
}

int kx_set_b(kx_t *kx, bn_t *b)
{
    return bn_copy(&kx->b, b);
}

int kx_get_sk(kx_t *kx, unsigned char *out, unsigned int n)
{
    if (n >= bn_length(&kx->Q))
        return FAILURE;

    if (bn_modexp(&kx->S, &kx->b, &kx->a, &kx->P) != SUCCESS)
        return FAILURE;

    bn_to_bytes(&kx->S, out, n / 8);
    return SUCCESS;
}

int kx_self_test()
{
    const static unsigned char test_a[] = {0x2e, 0x4e, 0x3a, 0x05, 0xd8, 0xff, 0x06, 0x2b, 0x4e, 0x46, 0xda, 0x6e, 0x12, 0x9c, 0xf3, 0x40, 0xb3, 0x07, 0x66, 0x99, 0x8a, 0x0e, 0x09, 0xe3, 0x93, 0x65, 0xa5, 0x52, 0x2d, 0x84, 0x83, 0x04},
        test_b[] = {0x6d, 0xd0, 0xae, 0x87, 0x4e, 0x86, 0xa8, 0x11, 0xf8, 0x1b, 0x2e, 0xf5, 0x25, 0x61, 0x50, 0x8d, 0x91, 0x4a, 0x43, 0x53, 0xb1, 0x24, 0x73, 0xe1, 0x18, 0xff, 0x35, 0xdb, 0x2c, 0x77, 0xbd, 0x4e, 0x24, 0x3f, 0xd8, 0x81, 0x82, 0xcd, 0x9e, 0x30, 0xf4, 0xd1, 0x62, 0xb1, 0xe5, 0x57, 0xd5, 0x09, 0x66, 0xfa, 0x9c, 0xdf, 0x97, 0xd4, 0xf6, 0x94, 0xc2, 0x68, 0xb8, 0x91, 0x7d, 0x40, 0xa5, 0x43, 0x96, 0xbe, 0xc4, 0xa9, 0x50, 0x93, 0xbf, 0x13, 0x44, 0x06, 0xaf, 0x3a, 0xb2, 0xf6, 0xf6, 0x22, 0x1f, 0x11, 0x38, 0xae, 0x7e, 0x6c, 0xa1, 0xb4, 0xb9, 0xd1, 0x07, 0x4f, 0x06, 0xf7, 0x09, 0x8d, 0x85, 0x11, 0x38, 0xd4, 0xd4, 0xb8, 0x8f, 0x4d, 0xe1, 0xce, 0x9f, 0x11, 0xb0, 0x49, 0xac, 0x5b, 0x74, 0xef, 0x8b, 0x8e, 0xa4, 0x80, 0xb3, 0xdf, 0xea, 0x85, 0x18, 0x58, 0xc0, 0x3f, 0x79, 0x31, 0x99, 0x96, 0x9f, 0x8e, 0x12, 0x08, 0xb2, 0x32, 0x3f, 0x73, 0xdb, 0x31, 0x4b, 0x19, 0x41, 0x06, 0xb9, 0x98, 0x4b, 0xe5, 0x13, 0xd8, 0xb8, 0xab, 0x4e, 0x1d, 0x0e, 0xff, 0x5a, 0x1a, 0x13, 0xa0, 0x9c, 0xa1, 0x74, 0x59, 0xce, 0x9a, 0x6e, 0x5e, 0xf5, 0xec, 0x2e, 0x3a, 0xcb, 0x76, 0x61, 0x33, 0x36, 0x59, 0x70, 0xa8, 0xcb, 0x86, 0x88, 0x24, 0xff, 0x14, 0x61, 0x8e, 0x0d, 0x2d, 0x58, 0xe2, 0x45, 0xb9, 0x8b, 0x1c, 0xb6, 0x66, 0xa7, 0xa0, 0x7d, 0xb1, 0x45, 0x93, 0x1b, 0xe6, 0xb0, 0x4c, 0xb2, 0xd4, 0xe2, 0x62, 0x0b, 0x26, 0x30, 0x91, 0x40, 0xe9, 0x1e, 0x60, 0x8e, 0xcd, 0xba, 0x19, 0x92, 0x1b, 0xdd, 0xf3, 0xd7, 0xff, 0x7b, 0x12, 0x97, 0x5c, 0x9c, 0xe9, 0x5e, 0x28, 0x82, 0x6d, 0xbc, 0xc9, 0xd8, 0x77, 0x47, 0x94, 0x04, 0xd6, 0xa7, 0x74, 0xcb, 0xc8, 0xfa, 0x78, 0x08, 0xa0},
        exp_A[] = {0x69, 0x72, 0x9a, 0x40, 0x98, 0xdc, 0x2c, 0x5c, 0xaa, 0xc5, 0xd8, 0x27, 0x9a, 0xc4, 0xa1, 0x2f, 0x24, 0x66, 0x8f, 0xd0, 0xe4, 0xf1, 0x17, 0xd3, 0x51, 0xa7, 0xb3, 0xcf, 0x00, 0x5c, 0x5e, 0x31, 0x74, 0x19, 0xca, 0xeb, 0xb1, 0x13, 0x1f, 0xd7, 0x58, 0x45, 0x29, 0x79, 0x21, 0x03, 0x88, 0x89, 0xcd, 0x31, 0x4d, 0xff, 0x9f, 0x56, 0xaf, 0xf2, 0x18, 0xcd, 0x81, 0x04, 0x0b, 0x17, 0x75, 0x63, 0x60, 0x1d, 0xfe, 0x58, 0xa0, 0xe0, 0x2b, 0x2f, 0xf3, 0x57, 0x46, 0xcb, 0xee, 0x51, 0x76, 0xcb, 0x4e, 0x69, 0x34, 0xcf, 0xa2, 0x7d, 0x59, 0xcc, 0x12, 0x08, 0x89, 0x76, 0x6b, 0x72, 0xcc, 0xe4, 0x6c, 0x09, 0x79, 0x57, 0xd0, 0x06, 0xe0, 0xf2, 0x21, 0xc3, 0xe2, 0x30, 0x3b, 0x63, 0x41, 0x6f, 0xe5, 0x84, 0xa3, 0x8a, 0x8b, 0x79, 0xb7, 0x92, 0xbc, 0xd5, 0xb6, 0xe7, 0xfe, 0xa2, 0x9a, 0x1d, 0x26, 0x00, 0x52, 0x92, 0x02, 0xb3, 0xdf, 0xf2, 0xf3, 0x41, 0x42, 0xa6, 0x82, 0xd9, 0xad, 0x0a, 0xf8, 0x32, 0x1c, 0x11, 0x14, 0x2c, 0x9d, 0xc3, 0x08, 0x6a, 0x92, 0x59, 0x3a, 0x61, 0x8a, 0xda, 0xc9, 0xbb, 0x46, 0x4f, 0xc8, 0x68, 0x3d, 0xee, 0x6a, 0xac, 0x15, 0xb0, 0x94, 0x64, 0x27, 0x74, 0xc9, 0xe4, 0xa8, 0xf6, 0xb6, 0x0e, 0x6c, 0x26, 0x0b, 0x31, 0x38, 0x9d, 0x40, 0xd7, 0x34, 0x3c, 0xd4, 0xe5, 0xf6, 0xea, 0xa0, 0xcb, 0xec, 0xfa, 0x92, 0x0a, 0x6a, 0x9d, 0x95, 0x52, 0x39, 0x22, 0x29, 0x54, 0x2d, 0x52, 0x43, 0x50, 0xab, 0x94, 0xf9, 0xda, 0x45, 0x6f, 0xa7, 0xd1, 0x64, 0x18, 0x25, 0x3e, 0xb4, 0x0a, 0xda, 0xf6, 0xb7, 0x49, 0x7d, 0xe2, 0x0c, 0x9c, 0x26, 0xed, 0xbb, 0xdc, 0x56, 0xc2, 0x7f, 0x43, 0x2e, 0xad, 0xac, 0x5e, 0x1f, 0xbc, 0x9f, 0xdf, 0x2c, 0xe8, 0x4b, 0xe6},
        exp_k[] = {0x25, 0xe1, 0xd6, 0xba, 0x1f, 0x49, 0x90, 0x47, 0x87, 0x87, 0xc8, 0x54, 0xfb, 0xed, 0xe8, 0xff, 0xe9, 0x5e, 0x2d, 0xc5, 0xed, 0xf7, 0x88, 0x36, 0x5f, 0x27, 0x98, 0x32, 0x7e, 0x10, 0x77};
    int result = FAILURE;
    kx_t kx;
    bn_t A, b, exp;
    unsigned char k[31];
    
    bn_init(&A);
    bn_init(&b);
    bn_init(&exp);

    if (kx_init_std(&kx, KX_GROUP_2048_256) != SUCCESS)
        goto fail2;

    if (bn_from_bytes(&kx.a, test_a, sizeof(test_a)) != SUCCESS)
        goto fail;

    if (kx_get_A(&kx, &A) != SUCCESS)
        goto fail;

    if (bn_from_bytes(&exp, exp_A, sizeof(exp_A)) != SUCCESS)
        goto fail;

    if (bn_cmp(&A, &exp) != 0)
        goto fail;

    if (bn_from_bytes(&b, test_b, sizeof(test_b)) != SUCCESS)
        goto fail;

    if (kx_set_b(&kx, &b) != SUCCESS)
        goto fail;

    if (kx_get_sk(&kx, k, sizeof(k) * 8) != SUCCESS)
        goto fail;

    if (memcmp(k, exp_k, sizeof(k)) != 0)
        goto fail;

    result = SUCCESS;

fail:
    kx_destroy(&kx);
fail2:
    if (result == FAILURE)
        fdprintf(STDERR, "KX self-test FAILED!\n");
    bn_destroy(&A);
    bn_destroy(&b);
    bn_destroy(&exp);
    return result;
}
